require 'rack/handler'
require 'stringio'

module Rack
  module Handler
    class Miniserver
      RACK_ERRORS       = "rack.errors"
      RACK_INPUT        = "rack.input"
      RACK_VERSION      = 'rack.version'
      RACK_MULTITHREAD  = 'rack.multithread'
      RACK_MULTIPROCESS = 'rack.multiprocess'
      RACK_RUN_ONCE     = 'rack.run_once'
      RACK_URL_SCHEME   = 'rack.url_scheme'

      DEFAULT_OPTIONS = {
        daemonize: false,
        Host: '0.0.0.0',
        Port: 8080,
        rackup: ::File.expand_path('config.ru'),
        log: ::File.expand_path('log/miniserver.log'),
        environment: 'development'
      }

      class << self
        def run(app, options={})
          options = DEFAULT_OPTIONS.merge(options)

          options[:host] = options.delete(:Host)
          options[:port] = options.delete(:Port)
          options[:daemonize] = false

          server = ::Miniserver::Server.new(app, options)

          yield(server) if block_given?

          server.start
        end

        def valid_options
          {
            "Host=HOST" => "Hostname to listen on (default: 0.0.0.0)",
            "Port=PORT" => "Port to listen on (default: 8080)",
          }
        end
      end

      def initialize(app)
        @app = app
      end

      def process(request, response)
        env = Hash.new

        request.headers.each_pair do |name, value|
          name = "#{::Miniserver::HTTP_PREFIX}#{name.upcase.tr(::Miniserver::MINUS, ::Miniserver::UNDERLINE)}"
          env[name] = value
        end

        %w(CONTENT_TYPE CONTENT_LENGTH).each do |name|
          env[name] = env.delete("HTTP_#{name}") if env["HTTP_#{name}"]
        end

        uri = URI(request.request_url)

        env[::Miniserver::REQUEST_METHOD] = request.http_method
        env[::Miniserver::REMOTE_ADDR]    = request.remote_addr
        env[::Miniserver::SERVER_NAME]    = request.server_name
        env[::Miniserver::SERVER_PORT]    = request.server_port
        env[::Miniserver::QUERY_STRING]   = request.query_string
        env[::Miniserver::REQUEST_PATH]   = uri.path
        env[::Miniserver::PATH_INFO]      = uri.path
        env[::Miniserver::SCRIPT_NAME]    = ::Miniserver::EMPTY_STRING
        env[::Miniserver::FRAGMENT]       = uri.fragment || ::Miniserver::EMPTY_STRING

        env[RACK_ERRORS]                  = STDERR
        env[RACK_INPUT]                   = StringIO.new(request.body)
        env[RACK_VERSION]                 = Rack::VERSION
        env[RACK_MULTITHREAD]             = false
        env[RACK_MULTIPROCESS]            = false
        env[RACK_RUN_ONCE]                = false
        env[RACK_URL_SCHEME]              = ::Miniserver::HTTP

        begin
          response.parse(*@app.call(env))
        ensure
          request.body.clear
        end
      end
    end

    register :miniserver, Miniserver
  end
end
